@@ -1,11 +1,8 @@
 #!/usr/bin python3
 
 """
-Coronavirus (COVID-19) Dashboard - API Service
-==============================================
-
-Software Development Kit (SDK)
-------------------------------
+Coronavirus Dashboard API: Python SDK
+=====================================
 
 This is a Python SDK for the COVID-19 API, as published by
 Public Health England on `Coronavirus (COVID-19) in the UK`_
@@ -33,7 +30,7 @@ __author__ = "Pouria Hadjibagheri"
 __copyright__ = "Copyright (c) 2020, Public Health England"
 __description__ = "SDK for the COVID-19 API (Coronavirus Dashboard in the UK)"
 __license__ = "MIT"
-__version__ = "1.1.4"
+__version__ = "1.2.2"
 __url__ = "https://coronavirus.data.gov.uk/"
 # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 
