@@ -15,6 +15,8 @@ import certifi
 
 # Internal:
 from uk_covid19.utils import save_data
+from uk_covid19.data_format import DataFormat
+from uk_covid19.exceptions import FailedRequestError
 
 # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 
@@ -29,53 +31,85 @@ FiltersType = Iterable[str]
 
 class Cov19API:
     """
-    COVID-19 API
-    ------------
     Interface to access the API service for COVID-19 data in the United Kingdom.
+
+    Parameters
+    ----------
+    filters: Iterable[str]
+        API filters. See the API documentations for additional
+        information.
+
+    structure: Dict[str, Union[dict, str]]
+        Structure parameter. See the API documentations for
+        additional information.
+
+    latest_by: Union[str, None]
+        Retrieves the latest value for a specific metric. [Default: ``None``]
     """
     endpoint = "https://api.coronavirus.data.gov.uk/v1/data"
-    _valid_formats = [
-        "csv",
-        "json",
-        "xml"
-    ]
+    release_timestamp_endpoint = "https://api.coronavirus.data.gov.uk/v1/timestamp"
+
     _last_update: Union[str, None] = None
-    total_pages: Union[int, None] = None
+    _total_pages: Union[int, None] = None
 
     def __init__(self, filters: FiltersType, structure: StructureType,
                  latest_by: Union[str, None] = None):
-        """
-        Parameters
-        ----------
-        filters: Iterable[str]
-            API filters. See the API documentations for additional
-            information.
-
-        structure: Dict[str, Union[dict, str]]
-            Structure parameter. See the API documentations for
-            additional information.
-
-        latest_by: Union[str, None]
-            Retrieves the latest value for a specific metric. [Default: ``None``]
-        """
         self.filters = filters
+
+        if any(isinstance(value, (list, dict)) for value in structure):
+            raise TypeError(
+                "Nested structures are no longer supported. Please define a flat "
+                "structure instead."
+            )
+
         self.structure = structure
         self.latest_by = latest_by
 
+    @property
+    def total_pages(self) -> Union[int, None]:
+        """
+        :property:
+            Produces the total number of pages for a given set of
+            parameters (only after the data are requested).
+
+        Returns
+        -------
+        Union[int, None]
+        """
+        return self._total_pages
+
     @property
     def last_update(self) -> str:
         """
-        Produces the timestamp for the last update in GMT.
+        :property:
+            Produces the timestamp for the last update in GMT.
+
+        This property supplies the API time - i.e. the time at which the data were
+        deployed to the database. Please note that there will always be a difference
+        between this time and the timestamp that is displayed on the website, which may
+        be accessed via the ``.get_release_timestamp()`` method. The website timestamp
+        signifies the time at which the data were release to the API, and by extension
+        the website.
+
+        .. note::
+
+            The output is extracted from the header and is accurate to
+            the second.
+            
+        .. warning::
+
+            The ISO-8601 standard requires a ``"Z"`` character to be added
+            to the end of the timestamp. This is a timezone feature and is
+            not recognised by Python's ``datetime`` library. It is, however,
+            most other libraries; e.g. ``pandas``. If you wish to parse the
+            timestamp using the the ``datetime`` library, make sure that you
+            remove the trailing ``"Z"`` character.
 
         Returns
         -------
         str
             Timestamp, formatted as ISO-8601.
 
-        .. note::
-            The output is extracted from the header and is accurate to
-            the second.
-
         Examples
         --------
         >>> filters = ["areaType=region"]
@@ -92,14 +126,6 @@ class Cov19API:
         >>> print(timestamp)
         2020-07-27T20:29:16.000000Z
 
-        .. warning::
-            The ISO-8601 standard requires a ``"Z"`` character to be added
-            to the end of the timestamp. This is a timezone feature and is
-            not recognised by Python's ``datetime`` library. It is, however,
-            most other libraries; e.g. ``pandas``. If you wish to parse the
-            timestamp using the the ``datetime`` library, make sure that you
-            remove the trailing ``"Z"`` character.
-
         >>> from datetime import datetime
         >>> parsed_timestamp = datetime.fromisoformat(timestamp.strip("Z"))
         >>> print(parsed_timestamp)
@@ -112,11 +138,62 @@ class Cov19API:
 
         return timestamp.isoformat() + ".000000Z"
 
+    @staticmethod
+    def get_release_timestamp() -> str:
+        """
+        :staticmethod:
+            Produces the website timestamp in GMT.
+
+        .. versionadded:: 1.2.0
+
+        This property supplies the website timestamp - i.e. the time at which the data
+        were released to the API and by extension the website. Please note that there
+        will be a difference between this timestamp and the timestamp produced using
+        the ``last_update`` property. The latter signifies the time at which the data
+        were deployed to the database, not the time at which they were released.
+
+        .. note::
+
+            The output is extracted from the header and is accurate to
+            the miliseconds.
+
+        .. warning::
+
+            The ISO-8601 standard requires a ``"Z"`` character to be added
+            to the end of the timestamp. This is a timezone feature and is
+            not recognised by Python's ``datetime`` library. It is, however,
+            most other libraries; e.g. ``pandas``. If you wish to parse the
+            timestamp using the the ``datetime`` library, make sure that you
+            remove the trailing ``"Z"`` character.
+            
+        Returns
+        -------
+        str
+            Timestamp, formatted as ISO-8601.
+
+        Examples
+        --------
+        >>> release_timestamp = Cov19API.get_release_timestamp()
+        >>> print(release_timestamp)
+        2020-08-08T15:00:09.977840Z
+
+        >>> from datetime import datetime
+        >>> release_timestamp = Cov19API.get_release_timestamp()
+        >>> parsed_timestamp = datetime.fromisoformat(release_timestamp.strip("Z"))
+        >>> print(parsed_timestamp)
+        2020-08-08 15:00:09
+        """
+        with request("GET", Cov19API.release_timestamp_endpoint) as response:
+            json_data = response.json()
+
+        return json_data['websiteTimestamp']
+
     @property
     def api_params(self) -> dict:
         """
-        API parameters, constructed based on ``filters``, ``structure``,
-        and ``latest_by`` arguments as defined by the user.
+        :staticmethod:
+            API parameters, constructed based on ``filters``, ``structure``,
+            and ``latest_by`` arguments as defined by the user.
 
         Returns
         -------
@@ -171,7 +248,8 @@ class Cov19API:
     @staticmethod
     def options():
         """
-        Provides the options by calling the ``OPTIONS`` method of the API.
+        :staticmethod:
+            Provides the options by calling the ``OPTIONS`` method of the API.
 
         Returns
         -------
@@ -194,7 +272,7 @@ class Cov19API:
             response.raise_for_status()
             return response.json()
 
-    def _get(self, format_as: str) -> Iterator[Response]:
+    def _get(self, format_as: DataFormat) -> Iterator[Response]:
         """
         Extracts paginated data by requesting all of the pages
         and combining the results.
@@ -202,38 +280,45 @@ class Cov19API:
         Parameters
         ----------
         format_as: str
-            Response format - Must be one of ``"json"``, ``"xml"``, or ``"csv"``.
+            Response format.
 
         Returns
         -------
         Iterator[Response]
-        """
-        if format_as not in self._valid_formats:
-            raise ValueError(
-                f"`format_as` must be one of '{str.join(', ', self._valid_formats)}'"
-            )
 
+        Raises
+        ------
+        FailedRequestError
+            When the request fails.
+        """
         api_params = self.api_params
 
         api_params.update({
-            "format": format_as,
+            "format": format_as.value,
             "page": 1
         })
 
+        if self.latest_by is not None:
+            del api_params["page"]
+
         while True:
             with request("GET", self.endpoint, params=api_params,
                          verify=certifi.where()) as response:
                 if response.status_code >= HTTPStatus.BAD_REQUEST:
-                    raise RuntimeError(f'Request failed: {response.text}')
-                if response.status_code == HTTPStatus.NO_CONTENT:
-                    self.total_pages = api_params["page"] - 1
+                    raise FailedRequestError(response=response, params=api_params)
+
+                if self.latest_by is not None:
+                    yield response
                     break
+                elif response.status_code == HTTPStatus.NO_CONTENT:
+                    self._total_pages = api_params["page"] - 1
+                    break
+                else:
+                    self._last_update = response.headers["Last-Modified"]
+                    yield response
 
-                self._last_update = response.headers["Last-Modified"]
-
-                yield response
-
-            api_params["page"] += 1
+            if self.latest_by is None:
+                api_params["page"] += 1
 
     def get_json(self, save_as: Union[str, None] = None,
                  as_string: bool = False) -> Union[dict, str]:
@@ -250,6 +335,8 @@ class Cov19API:
             extension -- i.e. ``.json`` for JSON).
 
         as_string: bool
+            .. versionadded:: 1.1.4
+
             If ``False`` (default), returns the data as a dictionary.
             Otherwise, returns the data as a JSON string.
 
@@ -277,7 +364,7 @@ class Cov19API:
             "data": list()
         }
 
-        for response in self._get("json"):
+        for response in self._get(DataFormat.JSON):
             current_data = response.json()
             page_data = current_data['data']
 
@@ -285,7 +372,7 @@ class Cov19API:
 
         resp["lastUpdate"] = self.last_update
         resp["length"] = len(resp["data"])
-        resp["totalPages"] = self.total_pages
+        resp["totalPages"] = self._total_pages
 
         if as_string:
             return dumps(resp, separators=(",", ":"))
@@ -294,7 +381,7 @@ class Cov19API:
             return resp
 
         data = dumps(resp, separators=(",", ":"))
-        save_data(data, save_as, "json")
+        save_data(data, save_as, DataFormat.JSON)
 
         return resp
 
@@ -312,6 +399,8 @@ class Cov19API:
             extension -- i.e. ``.xml`` for XML).
 
         as_string: bool
+            .. versionadded:: 1.1.4
+
             If ``False`` (default), returns an ``ElementTree``
             object. Otherwise, returns the data as an XML string.
 
@@ -347,7 +436,7 @@ class Cov19API:
 
         resp = XMLElement("document")
 
-        for response in self._get("xml"):
+        for response in self._get(DataFormat.XML):
             decoded_content = response.content.decode()
 
             # Parsing the XML:
@@ -361,7 +450,7 @@ class Cov19API:
         extras = {
             "lastUpdate": self.last_update,
             "length": len(resp.findall(".//data")),
-            "totalPages": self.total_pages
+            "totalPages": self._total_pages
         }
 
         for elm_name, value in extras.items():
@@ -378,7 +467,7 @@ class Cov19API:
         if as_string:
             return str_data
 
-        save_data(str_data, save_as, "xml")
+        save_data(str_data, save_as, DataFormat.XML)
 
         return resp
 
@@ -387,6 +476,7 @@ class Cov19API:
         Provides full data (all pages) in CSV.
 
         .. warning::
+
             Please make sure that the ``structure`` is not hierarchical as
             CSV outputs are defined as 2D tables and as such, do not support
             hierarchies.
@@ -404,6 +494,11 @@ class Cov19API:
         -------
         str
 
+        Raises
+        ------
+        ValueError
+            If the structure is nested.
+
         Examples
         --------
         >>> filters = ["areaType=region"]
@@ -437,7 +532,7 @@ class Cov19API:
         linebreak = "\n"
         resp = str()
 
-        for page_num, response in enumerate(self._get("csv"), start=1):
+        for page_num, response in enumerate(self._get(DataFormat.CSV), start=1):
             decoded_content = response.content.decode()
 
             # Removing CSV header (column names) where page
@@ -451,10 +546,43 @@ class Cov19API:
         if save_as is None:
             return resp
 
-        save_data(resp, save_as, "csv")
+        save_data(resp, save_as, DataFormat.CSV)
 
         return resp
 
+    def get_dataframe(self):
+        """
+        Provides the data as as ``pandas.DataFrame`` object.
+
+        .. versionadded:: 1.2.0
+
+        .. warning::
+
+            The ``pandas`` library is not included in the dependencies of this
+            library and must be installed separately.
+
+        Returns
+        -------
+        DataFrame
+
+        Raises
+        ------
+        ImportError
+            If the ``pandas`` library is not installed.
+        """
+        try:
+            from pandas import DataFrame
+        except ImportError:
+            raise ImportError(
+                "The `pandas` library is not installed as a part of the `uk-covid19` "
+                "library. Please install the library and try again."
+            )
+
+        data = self.get_json()
+        df = DataFrame(data["data"])
+
+        return df
+
     def __str__(self):
         resp = "COVID-19 in the UK - API Service\nCurrent parameters: \n"
         return resp + dumps(self.api_params, indent=4)
